// Apfelmnnchen in C++ fr beliebige Grafikauflsungen
// mit optionaler Bildabspeicherung als Windows(R)-Bitmap (*.BMP)
// (c) 1996 by Andreas Meile

#include <stdio.h>
#include <conio.h>
#include <math.h>
#include <grx.h>
#include "bmp_spei.h" // Paket mit Windows-Bitmap-Speicherfunktion
#include "eing_rou.h" // Paket mit syntaxprfenden Eingaberoutinen

// Prozedur fr Farbbergnge (lineares Verndern der RGB-Anteile)
void erzeuge_farbverlauf(int f1,  // Startfarbnummer
                         int r1,  // Start-Rotanteil
                         int g1,  // Start-Grnanteil
                         int b1,  // Start-Blauanteil
                         int f2,  // Endfarbnummer
                         int r2,  // End-Rotanteil
                         int g2,  // End-Grnanteil
                         int b2)  // End-Blauanteil
{
  const int dr = r2 - r1,
            dg = g2 - g1,
            db = b2 - b1,
            df = f2 - f1;
  for(register int i=f1; i<=f2; i++) {
    register int j=i - f1;
    GrSetColor(i, r1 + (j * dr) / df, g1 + (j * dg) / df, b1 + (j * db) / df);
  }
}

void main() {
  const float y_x_verhaeltnis = 3.0 / 4.0; // Bildschirmverhltnis: Bei
  // praktisch smtlichen heutigen Bildrhren immer 4:3, sodass in den
  // Auflsungen 640x480, 800x600 und 1024x768 quadratische Bildpunkte
  // resultieren

  int treiber_konst = GR_width_height_color_graphics, // Hardware-Konstante
  // fr Grafikaktivierung; falls mit lterem Treiber ohne Explizit-Modus-
  // angabe gearbeitet wird, muss vom Benutzer ein anderer Wert eingegeben
  // werden

      x_aufl = 640,  // horizonztale Bildauslsung (jeweils nur Vorgabewerte)
      y_aufl,        // veritkale Bildauflsung
      a_far = 16,    // Anzahl Farben
      a_far1,        // Hchste Farbnummer (fr Modulo-Operation)
      r_tiefe = 500, // Rechentiefe (Orbitlnge)
      vmodus_alt,    // Alter Videomodus zwischenspeichern
      ueb_jn,        // Option fr bersichtdarstellung
      echtf_flag;    // Flag, fr Echtfarbe : 0 = Palettenfarben,
                     // 1 = 32K-Farben, 2 = 24Bit-Echtfarbe

  const float rot_sinus   = 0.4,
              gruen_sinus = rot_sinus * sqrt(2.0),
              blau_sinus = rot_sinus * sqrt(3.0); // Absichtlich irrationale
              // Werte, um unschne Wiederholungen vermeiden zu knnen

  float xmin,      // Linke Ausschnittbegrenzung
        xmax,      // Rechte Ausschnittbegrenzung
        ymin,      // Untere Ausschnittbegrenzung
        ymax,      // Obere Ausschnittbegrenzung
        tmp_xmin,  // Temporre linke Ausschnittbegrenzung fr Eingabe
        tmp_xmax,  // Temporre rechte Ausschnittbegrenzung fr Eingabe
        tmp_ymin,  // Temporre untere Ausschnittbegrenzung fr Eingabe
        tmp_ymax,  // Temporre obere Ausschnittbegrenzung fr Eingabe
        xb,        // Basiskoordinate fr Pixel (0,0) horizontal
        yb,        // Basiskoordinate fr Pixel (0,0) vertikal
        xs,        // Schrittweite horizontal
        ys,        // Schrittweite vertikal
        z0r = 0.0, // Realteil der Bildkonstante z0 (Vorgabewert!)
        z0i = 0.0, // Imaginrteil der Bildkonstante z0
        z0r2,      // Quadrat des Realteils von z0
        z0i2,      // Quadrat des Imaginrteils von z0
        cr,        // Realteil der Konstante c
        ci;        // Imaginrteil der Konstante c

  // Diese Variablen mssen besonders schnell arbeiten knnen, denn
  // hier steckt der geschwindigkeitsentscheidende Algorithmus dahinter!
  register int anz_it, // Zhler fr die Iterationen
               x,      // Bildpunktzhler horizontal
               y;      // Bildpunktzhler vertikal

  register float r,  // Aktueller Realteil von z
                 i,  // Aktueller Imaginrteil von z
                 r2, // Quadrat von r
                 i2; // Quadrat von i

  char bildname[80]; // Bildname, falls Speicherung erwnscht

  // Eingabe der Auflsung
  printf("\e[2J\e[2;1HEinfaches Apfelmnnchen\n");
  printf("(c) 1996 by Andreas Meile\n\n");
  printf("Sie knnen sich die Grafikauflsung selber aussuchen!\n");
  
  printf("Wichtiger Hinweis: ltere Grafiktreiber knnen Sie nur ber die\n"
  "Treiberkonstante ansteuern, ndern Sie diese daher nur ab, falls auf\n"
  "Ihrem System die Direkt-Videomoduswahl nicht korrekt luft.\n\n");
  lies_interv_ganze_zahl("Grafiktreiberkonstante Ihres Systems", "/",
  GR_320_200_graphics, GR_width_height_color_graphics, &treiber_konst);
  
  lies_positive_zahl("Horizontale Auflsung (Anzahl Bildpunkte)", "/",
  &x_aufl);
  y_aufl = (int)(x_aufl * y_x_verhaeltnis + .9); // Addition von 0.9
  // nur, um Vorgabewert 0 zu vermeiden
  lies_positive_zahl("Vertikale Auflsung (Anzahl Bildpunkte)", "/",
  &y_aufl);
  lies_positive_zahl("Anzahl Farben (log2(Eingabe) ganzzahlig)", "/",
  &a_far);

  // Parameter fr Figur
  printf("\nWerte fr das Bild\n");
  lies_positive_zahl("Rechentiefe", "/", &r_tiefe);
  printf("Festlegung der Bildkonstante z0\n");
  lies_interv_fliesskommazahl("Realteil", "/", -2.0, 2.0, &z0r);
  z0i2 = sqrt(4.0 - z0r * z0r);
  lies_interv_fliesskommazahl("Imaginrteil", "/", -z0i2, z0i2, &z0i);
  
  // Ausschnitt bestimmen; dazu zuerst den bersichtsausschnitt berechnen
  tmp_xmin = z0i * z0i - z0r * z0r; // Mitte
  tmp_ymin = -2.0 * z0r * z0i;
  xmin = tmp_xmin - 2.0 / y_x_verhaeltnis;
  xmax = tmp_xmin + 2.0 / y_x_verhaeltnis;
  ymin = tmp_ymin - 2.0;
  ymax = tmp_ymin + 2.0;
  tmp_xmin = xmin - 0.5;
  tmp_xmax = xmax + 0.5;
  tmp_ymin = ymin - 0.5;
  tmp_ymax = ymax + 0.5;

  printf("Festlegung eines Ausschnittes fr c innerhalb der Gauss'schen "
  "Ebene,\ndie Vorgabewerte entsprechen dabei dem bersichtsausschnitt\n");

  ueb_jn=lies_interv_fliesskommazahl("Linke Begrenzung (=bersicht)",
  "///", tmp_xmin, tmp_xmax, &xmin);
  if (ueb_jn >= 1) { // bersicht erwnscht
    xmin = xmin - 0.02 / y_x_verhaeltnis;
    xmax = xmax + 0.02 / y_x_verhaeltnis;
    ymin = ymin - 0.02;
    ymax = ymax + 0.02;
  } else {
    lies_interv_fliesskommazahl("Rechte Begrenzung", "/", xmin, tmp_xmax,
    &xmax);
    lies_interv_fliesskommazahl("Untere Begrenzung", "/", tmp_ymin, tmp_ymax,
    &ymin);
    // Vorgabewert ausproportionieren
    ymax = ymin + y_x_verhaeltnis * (xmax - xmin);
    printf("Vorgabewert so gewhlt, dass eine verzerrungsfreie Darstellung"
    " entsteht.\n");
    lies_interv_fliesskommazahl("Obere Begrenzung", "/", ymin, tmp_ymax,
    &ymax);
  }

  printf("\nDurch Eingabe eines Dateinamen kann das Bild auf Diskette/Fest-\n"
         "platte als Windows(R) Bitmap abgespeichert werden, sodass es\n"
         "unter Microsoft Windows weiterverarbeitet werden kann.\n"
         "Falls Sie dies nicht wnschen, einfach <RETURN> alleine drcken\n");
  printf("\nBilddateiname (ohne .BMP!) oder Leereingabe : ");
  gets(bildname);

  // Alter Videomodus sichern
  vmodus_alt = GrCurrentMode();

  // Start der Berechnung
  GrSetMode(treiber_konst, x_aufl, y_aufl, a_far);
  // GRX sucht sich in der Regel selber den am nchsten an diese Daten
  // kommenden Videomodus aus, daher mssen die Werte mit den effektiv
  // gltigen Werten aktualisiert werden
  x_aufl = GrScreenX();
  y_aufl = GrScreenY();
  a_far = GrNumColors();
  
  echtf_flag = (a_far >= 32768) + (a_far >= 0x1000000); // Flag fr Punkt-
  // setzmethode setzen
  
  a_far1 = a_far;
  a_far1--;

  // Farbpaletten festlegen
  switch(a_far) {
    case 256 : {
      const int farb256_tabelle[22][4] = {{  0,  5,251,198},
        { 21,  3,240, 43}, { 28,  6,187, 23}, { 47,  2, 49,  7},
        { 53, 24, 17, 34}, { 75, 98,  8,178}, { 93,215, 27,243},
        {101,247, 72,242}, {117,254,196,250}, {126,246,251,211},
        {139,233,249,122}, {149,229,201, 65}, {162,208,115, 11},
        {167,179,105,  0}, {175,118,110, 15}, {192, 20, 80, 90},
        {203, 30, 61,170}, {216,100, 72,230}, {226,120,145,240},
        {239, 95,237,250}, {246, 41,250,251}, {255,  5,251,198}};
      // Zuweisen
      for(register int i = 0; i < 21; i++) {
        register int j = i+1;
        erzeuge_farbverlauf(farb256_tabelle[i][0],  // Startfarbnummer
                            farb256_tabelle[i][1],  // Start-Rotanteil
                            farb256_tabelle[i][2],  // Start-Grnanteil
                            farb256_tabelle[i][3],  // Start-Blauanteil
                            farb256_tabelle[j][0],  // Endfarbnummer
                            farb256_tabelle[j][1],  // End-Rotanteil
                            farb256_tabelle[j][2],  // End-Grnanteil
                            farb256_tabelle[j][3]); // End-Blauanteil
      }
      GrSetColor(255, 255, 0, 0); // Endgrenzfarbe = knallrot
      break;
    }
    case 16 : {
      const int farb16_tabelle[16][3] = {{193, 26, 93},
           {219, 76, 65}, {215,119, 33}, {223,171, 40},
           {223,213, 47}, {158,200, 57}, { 90,187, 75},
           { 58,167,101}, { 18,139,152}, {  6, 96,204},
           {  8, 57,235}, { 58, 19,215}, { 98,  7,192},
           {149,  9,159}, {180, 26,127}, {  6, 14, 12}};
      // Zuweisen
      for(int i=0; i<16; i++)
        GrSetColor(i, farb16_tabelle[i][0], farb16_tabelle[i][1],
                      farb16_tabelle[i][2]);
    }
  }

  // Basis und Schrittwerte festlegen
  xs = (xmax - xmin) / x_aufl;
  ys = (ymax - ymin) / y_aufl;
  xb = xmin + xs / 2.0;
  yb = ymax - ys / 2.0;
  z0r2 = z0r * z0r;
  z0i2 = z0i * z0i;

  for (y = 0; y < y_aufl && !kbhit(); y++) {
    ci = yb - y * ys;
    for (x = 0; x < x_aufl; x++) {
      cr = xb + x * xs;
      r = z0r;
      i = z0i;
      r2 = z0r2;
      i2 = z0i2;
      anz_it = 0;
      while(anz_it < r_tiefe && r2 + i2 < 4) {
        i = 2.0 * r * i + ci;
        r = r2 - i2 + cr;
        anz_it++;
        r2 = r * r;
        i2 = i * i;
      }
      if(echtf_flag) {
        register int rot, gruen, blau;
        if(anz_it == r_tiefe) {
          rot = 0;
          gruen = 0;
          blau = 0;
        } else {
          register float w=(float)anz_it;
          rot = (int)(127.999 * sin(w * rot_sinus) + 128.0);
          gruen = (int)(127.999 * sin(w * gruen_sinus) + 128.0);
          blau = (int)(127.999 * cos(w * blau_sinus) + 128.0); // cos(),
          // damit beim Start (Iterationen = 0) ein dezentes Blau entsteht
        }
        if(echtf_flag == 2) // 24Bit-Echtfarbe
          GrPlot(x, y, rot << 16 | gruen << 8 | blau);
        else
          GrPlot(x, y, (rot & 0xf8) << 7 | (gruen & 0xf8) << 2 | blau >> 3);
      } else {
        if(anz_it == r_tiefe)
          GrPlot(x, y, a_far1);
        else
          GrPlot(x, y, anz_it % a_far1);
      }
    }
  }

  // Bild speichern, fall erwnscht
  int erfolg; // fr den Rckgabewert der Bildabspeicherungsprozedur
  {
    register int i=-1;
    while(bildname[++i]); // Zeichenkettenende suchen
    if(i) { // Falls nicht Null, dann wurde Speicherung erwnscht
      bildname[i++] = '.';
      bildname[i++] = 'B';
      bildname[i++] = 'M';
      bildname[i++] = 'P';
      bildname[i] = '\0'; // Stringabschluss-Nullbyte
      erfolg = speichere_als_bmp(bildname);
    }
  }

  // Warten auf Tastendruck
  getch();
  
  // Wieder in den alten Videomodus (in der Regel Textmodus) zurckschalten
  GrSetMode(vmodus_alt);
  printf("\e[2J\e[2;1HDie vorherige Grafik bestand aus folgenden Daten:\n\n");
  printf("Bildschirmauflsung: %dx%d Bildpunkte mit %d Farben.\n\n", x_aufl,
  y_aufl, a_far);
  printf("Bildkonstante: z0 = %f %+f * j\n", z0r, z0i);
  printf("Rechentiefe: Abbruch nach %d Iterationen\n\n", r_tiefe);
  printf("Ausschnitt aus der Gauss'schen Ebene\n");
  printf("Linke, untere Ecke: %f %+f * j\n", xmin, ymin);
  printf("Rechte, obere Ecke: %f %+f * j\n", xmax, ymax);
  printf("Resultierendes Bildverhltnis x:y = %f\n\n", (xmax - xmin) / (ymax
  - ymin) * y_x_verhaeltnis);
  if(bildname[0])
    switch(erfolg) {
      case 0 : {
        printf("Bild erfolgreich unter %s abgespeichert\n", bildname);
        break;
      }
      case 1 : {
        printf("Fehler whrend des Abspeicherns unter %s\n", bildname);
        printf("Bild unvollstndig gespeichert\n");
        break;
      }
      case 2 : {
        printf("Fehler beim Anlegen der Datei %s\n", bildname);
        printf("(mglicherweise Fehler im Pfad) => nichts gespeichert\n");
        break;
      }
      default : {
        printf("Ein interner Fehler ist aufgetreten. Fehlercode: %d\n",
        erfolg);
      }
    }
  else
    printf("Bildabspeicherung wurde nicht erwnscht\n");
}
